# ruff: noqa: F401

# versioninfo.py will be installed by win32more-Microsoft.WindowsAppSDK
from .versioninfo import (  # noqa
    WINDOWSAPPSDK_RELEASE_MAJORMINOR,
    WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W,
    WINDOWSAPPSDK_RUNTIME_VERSION_UINT64,
)

from ._runtime import initialize, is_self_contained, get_loaded_appsdk_info, RuntimeNotFoundError



from __future__ import annotations

from win32more import FAILED, ComError, windll
from win32more.Windows.Win32.Foundation import APPMODEL_ERROR_NO_PACKAGE, STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED
from win32more.Windows.Win32.Storage.Packaging.Appx import PACKAGE_INFO, PACKAGE_VERSION

from . import (
    WINDOWSAPPSDK_RELEASE_MAJORMINOR,
    WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W,
    WINDOWSAPPSDK_RUNTIME_VERSION_UINT64,
)
from ._packaging import GetCurrentPackageInfo, GetFrameworkPackageFamilyName
from .mddbootstrap import MddBootstrapInitialize, _OnNoMatch_ShowUI


class RuntimeNotFoundError(OSError):
    def __init__(self) -> None:
        major = WINDOWSAPPSDK_RELEASE_MAJORMINOR >> 16
        minor = WINDOWSAPPSDK_RELEASE_MAJORMINOR & 0xFFFF
        min_version = PACKAGE_VERSION(Version=WINDOWSAPPSDK_RUNTIME_VERSION_UINT64)
        msg = f"This application requires the Windows App Runtime Version {major}.{minor} (MSIX package version >= {min_version.Major}.{min_version.Minor}.{min_version.Build}.{min_version.Revision})"
        super().__init__(STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED, msg)

    def show_dialog(self) -> None:
        _OnNoMatch_ShowUI(
            WINDOWSAPPSDK_RELEASE_MAJORMINOR,
            WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W,
            PACKAGE_VERSION(Version=WINDOWSAPPSDK_RUNTIME_VERSION_UINT64),
        )


def initialize() -> None:
    if _appsdk_runtime_dll_exists():
        # case 1: This is packaged application and appsdk is statically loaded via AppxManifest.xml settings.
        # case 2: Application is self-contained package.
        return

    hr = MddBootstrapInitialize(
        WINDOWSAPPSDK_RELEASE_MAJORMINOR,
        WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W,
        PACKAGE_VERSION(Version=WINDOWSAPPSDK_RUNTIME_VERSION_UINT64),
    )

    if FAILED(hr):
        if hr == STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED:
            raise RuntimeNotFoundError()
        raise ComError(hr)


def get_loaded_appsdk_info() -> PACKAGE_INFO:
    family_name = GetFrameworkPackageFamilyName(
        WINDOWSAPPSDK_RELEASE_MAJORMINOR, WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W
    )
    for info in GetCurrentPackageInfo():
        if info.packageFamilyName == family_name:
            return info
    raise LookupError(f"{family_name} is not loaded")


# WindowsAppRuntime_IsSelfContained()
def is_self_contained() -> bool:
    try:
        get_loaded_appsdk_info()
    except LookupError:
        return True
    except OSError as e:
        if e.winerror == APPMODEL_ERROR_NO_PACKAGE:
            # Process is not packaged process and appsdk is not loaded dynamically.
            return True
        raise
    return False


def _appsdk_runtime_dll_exists() -> bool:
    try:
        windll["Microsoft.WindowsAppRuntime.dll"]
    # FileNotFoundError <= 3.12
    # AttributeError >= 3.13
    except (FileNotFoundError, AttributeError):
        return False
    return True



from __future__ import annotations

from ctypes import pointer

from win32more import FAILED, Byte, Char, String, UInt32, WinError
from win32more._win32 import ARCH
from win32more.Windows.Win32.Foundation import (
    APPMODEL_ERROR_NO_PACKAGE,
    ERROR_INSUFFICIENT_BUFFER,
    ERROR_SUCCESS,
)
from win32more.Windows.Win32.Storage.Packaging.Appx import (
    PACKAGE_FILTER_DYNAMIC,
    PACKAGE_FILTER_STATIC,
    PACKAGE_INFO,
    PACKAGE_VERSION,
    GetCurrentPackageFullName,
)
from win32more.Windows.Win32.Storage.Packaging.Appx import GetCurrentPackageInfo as _GetCurrentPackageInfo
from win32more.Windows.Win32.Storage.Packaging.Appx import GetPackagePathByFullName as _GetPackagePathByFullName
from win32more.Windows.Win32.Storage.Packaging.Appx import GetPackagesByPackageFamily as _GetPackagesByPackageFamily


def IsPackagedProcess() -> bool:
    return GetCurrentPackageFullName(UInt32(), None) != APPMODEL_ERROR_NO_PACKAGE


# https://learn.microsoft.com/en-us/windows/apps/desktop/modernize/package-identity-overview
def GetFrameworkPackageFamilyName(major_minor_version: int, version_tag: PACKAGE_VERSION) -> str:
    major_version = major_minor_version >> 16
    minor_version = major_minor_version & 0xFFFF
    version_tag_delimiter = "" if version_tag == "" else "-"
    name = f"Microsoft.WindowsAppRuntime.{major_version}.{minor_version}{version_tag_delimiter}{version_tag}"
    publisher_id = "8wekyb3d8bbwe"
    return f"{name}_{publisher_id}"


def GetFrameworkPackageDirectory(
    major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION
) -> str | None:
    family_name = GetFrameworkPackageFamilyName(major_minor_version, version_tag)
    full_name = ResolvePackageDependency(family_name, min_version, ARCH)
    if full_name is None:
        return None
    return GetPackagePathByFullName(full_name)


def ResolvePackageDependency(family_name: str, min_version: PACKAGE_VERSION, arch: str) -> str | None:
    bestfit = None
    latest_version = None

    for full_name in GetPackagesByPackageFamily(family_name):
        name, version, architecture, resource_id, publisher_id = full_name.split("_")
        major, minor, build, revision = [int(x) for x in version.split(".")]
        if architecture.upper() != arch:
            continue
        if (major, minor, build, revision) < (
            min_version.Major,
            min_version.Minor,
            min_version.Build,
            min_version.Revision,
        ):
            continue
        if latest_version is None or latest_version < (major, minor, build, revision):
            bestfit = full_name
            latest_version = (major, minor, build, revision)

    return bestfit


def GetPackagesByPackageFamily(family_name: str) -> list[str]:
    count = UInt32()
    buffer_length = UInt32()
    r = _GetPackagesByPackageFamily(family_name, count, None, buffer_length, None)
    if r == ERROR_SUCCESS:
        return []  # not found
    elif r != ERROR_INSUFFICIENT_BUFFER:
        raise WinError(r)

    full_names = (String * count.value)()
    buffer = (Char * buffer_length.value)()
    r = _GetPackagesByPackageFamily(family_name, count, full_names, buffer_length, buffer)
    if r != ERROR_SUCCESS:
        raise WinError(r)

    return full_names[: count.value]


def GetPackagePathByFullName(full_name: str) -> str:
    path_length = UInt32()
    r = _GetPackagePathByFullName(full_name, path_length, None)
    if r != ERROR_INSUFFICIENT_BUFFER:
        raise WinError(r)

    path = (Char * path_length.value)()
    r = _GetPackagePathByFullName(full_name, path_length, path)
    if r != ERROR_SUCCESS:
        raise WinError(r)

    return path.value


def GetCurrentPackageInfo(flags=PACKAGE_FILTER_DYNAMIC | PACKAGE_FILTER_STATIC) -> list[PACKAGE_INFO]:
    size = UInt32()
    r = _GetCurrentPackageInfo(flags, size, None, None)
    if r != ERROR_INSUFFICIENT_BUFFER:
        raise WinError(r)

    buf = bytearray(size.value)
    pinfo = pointer(PACKAGE_INFO.from_buffer(buf))
    count = UInt32()
    r = _GetCurrentPackageInfo(flags, size, Byte.from_buffer(buf), count)
    if FAILED(r):
        raise WinError(r)

    return pinfo[0 : count.value]



from __future__ import annotations

import os
import sys
import webbrowser
from pathlib import Path

import win32more
from win32more import FAILED, Int32, String, UInt32, Void
from win32more._win32 import ARCH, winfunctype
from win32more.Windows.Win32.Foundation import (
    HRESULT,
    S_OK,
    STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED,
)
from win32more.Windows.Win32.Storage.Packaging.Appx import (
    PACKAGE_DEPENDENCY_RANK_DEFAULT,
    PACKAGE_VERSION,
    PACKAGEDEPENDENCY_CONTEXT,
    AddPackageDependency,
    AddPackageDependencyOptions_None,
    CreatePackageDependencyOptions_None,
    PackageDependencyLifetimeKind_Process,
    PackageDependencyProcessorArchitectures_Arm64,
    PackageDependencyProcessorArchitectures_X64,
    PackageDependencyProcessorArchitectures_X86,
    TryCreatePackageDependency,
)
from win32more.Windows.Win32.UI.WindowsAndMessaging import IDYES, MB_ICONERROR, MB_YESNO, MessageBox

from . import WINDOWSAPPSDK_RUNTIME_VERSION_UINT64
from ._packaging import GetFrameworkPackageFamilyName, IsPackagedProcess

if ARCH == "ARM64":
    PackageDependencyProcessorArchitectures_Current = PackageDependencyProcessorArchitectures_Arm64
    os.add_dll_directory(os.path.dirname(win32more.__file__) + "\\dll\\arm64")
elif ARCH == "X64":
    PackageDependencyProcessorArchitectures_Current = PackageDependencyProcessorArchitectures_X64
    os.add_dll_directory(os.path.dirname(win32more.__file__) + "\\dll\\x64")
elif ARCH == "X86":
    PackageDependencyProcessorArchitectures_Current = PackageDependencyProcessorArchitectures_X86
    os.add_dll_directory(os.path.dirname(win32more.__file__) + "\\dll\\x86")
else:
    assert False

MddBootstrapInitializeOptions = Int32
MddBootstrapInitializeOptions_None = 0
MddBootstrapInitializeOptions_OnError_DebugBreak = 1
MddBootstrapInitializeOptions_OnError_DebugBreak_IfDebuggerAttached = 2
MddBootstrapInitializeOptions_OnError_FailFast = 4
MddBootstrapInitializeOptions_OnNoMatch_ShowUI = 8
MddBootstrapInitializeOptions_OnPackageIdentity_NOOP = 16


@winfunctype("Microsoft.WindowsAppRuntime.Bootstrap.dll", entry_point="MddBootstrapInitialize")
def _MddBootstrapInitialize(majorMinorVersion: UInt32, versionTag: String, minVersion: PACKAGE_VERSION) -> HRESULT: ...


@winfunctype("Microsoft.WindowsAppRuntime.Bootstrap.dll", entry_point="MddBootstrapInitialize2")
def _MddBootstrapInitialize2(
    majorMinorVersion: UInt32, versionTag: String, minVersion: PACKAGE_VERSION, options: MddBootstrapInitializeOptions
) -> HRESULT: ...


@winfunctype("Microsoft.WindowsAppRuntime.Bootstrap.dll", entry_point="MddBootstrapShutdown")
def _MddBootstrapShutdown() -> Void: ...


def MddBootstrapInitialize(major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION) -> int:
    return MddBootstrapInitialize2(major_minor_version, version_tag, min_version, MddBootstrapInitializeOptions_None)


def MddBootstrapInitialize2(
    major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION, options: int
) -> int:
    if IsPackagedProcess():
        # FIXME: Mddbootstrap API doesn't support packaged process.
        # Since 1.6.5, WindowsAppSDK seems to work with Win11's package dependency API.
        # I'm not sure if it is offically supported.  It might have a problem.
        VERSION_1_6_5 = 0x1770019109300000
        if not _IsWin11() or WINDOWSAPPSDK_RUNTIME_VERSION_UINT64 < VERSION_1_6_5:
            raise RuntimeError("Packaged process is not supported before Windows 11")
        hr = _Initialize_Win11(GetFrameworkPackageFamilyName(major_minor_version, version_tag), min_version)
        if FAILED(hr):
            if hr == STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED:
                if options & MddBootstrapInitializeOptions_OnNoMatch_ShowUI:
                    _OnNoMatch_ShowUI(major_minor_version, version_tag, min_version)
            return hr
        return S_OK
    else:
        return _MddBootstrapInitialize2(major_minor_version, version_tag, min_version, options)


def MddBootstrapShutdown() -> None:
    if IsPackagedProcess() and not _IsWin11():
        raise RuntimeError("Packaged process is not supported before Windows 11")
    elif _IsWin11():
        pass
    else:
        _MddBootstrapShutdown()


def _IsWin11() -> bool:
    return sys.getwindowsversion() >= (10, 0, 22000)


def _Initialize_Win11(family_name: str, min_version: PACKAGE_VERSION) -> int:
    package_dependency_id = String()
    hr = TryCreatePackageDependency(
        None,
        family_name,
        min_version,
        PackageDependencyProcessorArchitectures_Current,
        PackageDependencyLifetimeKind_Process,
        None,
        CreatePackageDependencyOptions_None,
        package_dependency_id,
    )
    if FAILED(hr):
        return hr

    package_dependency_context = PACKAGEDEPENDENCY_CONTEXT()
    hr = AddPackageDependency(
        package_dependency_id,
        PACKAGE_DEPENDENCY_RANK_DEFAULT,
        AddPackageDependencyOptions_None,
        package_dependency_context,
        None,
    )
    if FAILED(hr):
        return hr

    return S_OK


def _OnNoMatch_ShowUI(major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION) -> None:
    caption = f"{Path(sys.executable).name} - This application could not be started"
    text = f"""This application requires the Windows App Runtime
Version {major_minor_version >> 16}.{major_minor_version & 0xFFFF}
(MSIX package version >= {min_version.Major}.{min_version.Minor}.{min_version.Build}.{min_version.Revision})

Do you want to install a compatible Windows App Runtime now?"""
    r = MessageBox(0, text, caption, MB_YESNO | MB_ICONERROR)
    if r == IDYES:
        webbrowser.open("https://docs.microsoft.com/windows/apps/windows-app-sdk/downloads")
